// Copyright (C) 2012  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_GRAPH_CUTs_HEADER_
#define DLIB_GRAPH_CUTs_HEADER_

#include "graph_cuts/min_cut.h"
#include "graph_cuts/general_flow_graph.h"
#include "graph_cuts/find_max_factor_graph_potts.h"
#include "graph_cuts/graph_labeler.h"

#endif // DLIB_GRAPH_CUTs_HEADER_



