// Copyright (C) 2008  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.

#ifdef DLIB_ALL_SOURCE_END
#include "dlib_basic_cpp_build_tutorial.txt"
#endif

#ifndef DLIB_SVm_THREADED_HEADER
#define DLIB_SVm_THREADED_HEADER

#include "svm.h"
#include "svm/svm_threaded.h"
#include "svm/structural_svm_problem_threaded.h"
#include "svm/structural_svm_distributed.h"
#include "svm/structural_svm_object_detection_problem.h"
#include "svm/structural_object_detection_trainer.h"
#include "svm/structural_svm_sequence_labeling_problem.h"
#include "svm/structural_sequence_labeling_trainer.h"

#include "svm/structural_svm_assignment_problem.h"
#include "svm/structural_assignment_trainer.h"
#include "svm/cross_validate_track_association_trainer.h"
#include "svm/structural_track_association_trainer.h"

#include "svm/structural_svm_graph_labeling_problem.h"
#include "svm/structural_graph_labeling_trainer.h"
#include "svm/cross_validate_graph_labeling_trainer.h"
#include "svm/svm_multiclass_linear_trainer.h"
#include "svm/one_vs_one_trainer.h"
#include "svm/one_vs_all_trainer.h"
#include "svm/structural_sequence_segmentation_trainer.h"

#endif // DLIB_SVm_THREADED_HEADER



